@@ -13,6 +13,7 @@ import numpy as np
 import pandas as pd
 import torch
 import torch.optim as optim
+from loguru import logger
 from scipy.stats import norm
 
 from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
@@ -25,6 +26,7 @@ from libinephany.utils import optim_utils
 # ======================================================================================================================
 
 EXP_AVERAGE = "exp_avg"
+MOMENTUM_BUFFER = "momentum_buffer"
 MIN_DECAY_FACTOR = 1e-10
 
 MIN_TOTAL_WEIGHT = 1e-15  # Minimum total weight threshold for numerical stability
@@ -59,18 +61,54 @@ class StatisticStorageTypes(Enum):
 # ======================================================================================================================
 
 
-def get_exponential_weighted_average(values: list[int | float]) -> float:
+def get_exponential_weighted_average(
+    values: list[int | float], invalid_value_threshold: float = 1e10, tracker_name: str = "Unknown"
+) -> float:
     """
     :param values: List of values to average via EWA.
+    :param invalid_value_threshold: Threshold for invalid observations, default is 1e10.
+    :param tracker_name: Name of the tracker for error reporting, default is "Unknown".
     :return: EWA of the given values.
     """
+
+    # Logging statistic tracker name when input is invalid.
+
+    if len(values) == 0:
+        raise ValueError(
+            f"Statistic Tracker: {tracker_name} gathered data with empty list! It likely means a bug is triggered in the"
+            f" code related to {tracker_name} or the inner task. Please check up the code!"
+        )
+
+    if any(not isinstance(value, (float, int)) for value in values):
+        # Should never happen, but just in case
+        raise ValueError(
+            f"Statistic Tracker: {tracker_name} gathered data with invalid values (not int or float)! It likely means a"
+            f" bug is triggered in the code related to {tracker_name} or the inner task. Please check up the code!"
+        )
+
+    if any(abs(value) > invalid_value_threshold for value in values):
+        # check for large values (including negative values)
+        logger.warning(
+            f"Statistic Tracker: {tracker_name} gathered data with values out of range (invalid_value_threshold: "
+            f"{invalid_value_threshold})! May cause episode termination if StopEpisodeFromInvalidObservations is used."
+        )
+
+    if any(np.isnan(value) for value in values):
+        logger.warning(
+            f"Statistic Tracker: {tracker_name} gathered data with NaN values! May cause episode termination if "
+            f"StopEpisodeFromInvalidObservations is used."
+        )
+
     exp_weighted_average = pd.Series(values).ewm(alpha=0.1).mean().iloc[-1]
     assert isinstance(exp_weighted_average, float)
     return exp_weighted_average
 
 
 def apply_averaging_function_to_tensor_statistics(
-    tensor_statistics: list[TensorStatistics], averaging_function: Callable[[list[float]], float]
+    tensor_statistics: list[TensorStatistics],
+    averaging_function: Callable[[list[float], float, str], float],
+    invalid_value_threshold: float = 1e10,
+    tracker_name: str = "Unknown",
 ) -> TensorStatistics:
     """
     :param tensor_statistics: List of statistics models to average over.
@@ -80,40 +118,60 @@ def apply_averaging_function_to_tensor_statistics(
 
     fields = TensorStatistics.model_fields.keys()
     averaged_metrics = {
-        field: averaging_function([getattr(statistics, field) for statistics in tensor_statistics]) for field in fields
+        field: averaging_function(
+            [getattr(statistics, field) for statistics in tensor_statistics],
+            invalid_value_threshold,
+            tracker_name,
+        )
+        for field in fields
     }
 
     return TensorStatistics(**averaged_metrics)
 
 
 def apply_averaging_function_to_dictionary_of_tensor_statistics(
-    data: dict[str, list[TensorStatistics]], averaging_function: Callable[[list[float]], float]
+    data: dict[str, list[TensorStatistics]],
+    averaging_function: Callable[[list[float], float, str], float],
+    invalid_value_threshold: float = 1e10,
+    tracker_name: str = "Unknown",
 ) -> dict[str, TensorStatistics]:
     """
     :param data: Dictionary mapping parameter group names to list of TensorStatistics from that parameter group.
     :param averaging_function: Function to average the values with.
+    :param invalid_value_threshold: Threshold for invalid observations.
+    :param tracker_name: Name of the tracker for error reporting.
     :return: Dictionary mapping parameter group names to TensorStatistics averaged over all statistics in the given
     TensorStatistics models.
     """
 
     return {
         group: apply_averaging_function_to_tensor_statistics(
-            tensor_statistics=metrics, averaging_function=averaging_function
+            tensor_statistics=metrics,
+            averaging_function=averaging_function,
+            invalid_value_threshold=invalid_value_threshold,
+            tracker_name=tracker_name,
         )
         for group, metrics in data.items()
     }
 
 
 def apply_averaging_function_to_dictionary_of_metric_lists(
-    data: dict[str, list[float]], averaging_function: Callable[[list[float]], float]
+    data: dict[str, list[float]],
+    averaging_function: Callable[[list[float], float, str], float],
+    invalid_value_threshold: float = 1e10,
+    tracker_name: str = "Unknown",
 ) -> dict[str, float]:
     """
     :param data: Dictionary mapping parameter group names to list of metrics from that parameter group.
     :param averaging_function: Function to average the values with.
+    :param invalid_value_threshold: Threshold for invalid observations.
+    :param tracker_name: Name of the tracker for error reporting.
     :return: Dictionary mapping parameter group names to averages over all metrics from each parameter group.
     """
 
-    return {group: averaging_function(metrics) for group, metrics in data.items()}
+    return {
+        group: averaging_function(metrics, invalid_value_threshold, tracker_name) for group, metrics in data.items()
+    }
 
 
 def average_tensor_statistics(tensor_statistics: list[TensorStatistics]) -> TensorStatistics:
@@ -230,6 +288,24 @@ def form_update_tensor(
         raise NotImplementedError(f"Optimizer {type(optimizer).__name__} is not supported!")
 
 
+def form_momentum_tensor(
+    optimizer: optim.Optimizer, parameters: list[torch.Tensor], parameter_group: dict[str, Any]
+) -> None | torch.Tensor:
+    """
+    :param optimizer: Optimizer to form the momentum tensor from.
+    :param parameters: Parameters to create the momentum tensor from.
+    :param parameter_group: Parameter group within the optimizer the given parameters came from.
+    """
+    if type(optimizer) in optim_utils.ADAM_OPTIMISERS:
+        momentum_list = [optimizer.state[p][EXP_AVERAGE].view(-1) for p in parameters if tensor_on_local_rank(p)]
+        return torch.cat(momentum_list) if momentum_list else None
+    elif type(optimizer) in optim_utils.SGD_OPTIMISERS:
+        momentum_list = [optimizer.state[p][MOMENTUM_BUFFER].view(-1) for p in parameters if tensor_on_local_rank(p)]
+        return torch.cat(momentum_list) if momentum_list else None
+    else:
+        raise NotImplementedError(f"Optimizer {type(optimizer).__name__} is not supported!")
+
+
 def null_standardizer(value_to_standardize: float, **kwargs) -> float:
     """
     :param value_to_standardize: Value to mock the standardization of.
