@@ -24,6 +24,7 @@ from .hyperparameter_observers import (
     OptimizerTypeOneHot,
 )
 from .loss_observers import (
+    BestObservedValidationLoss,
     LHOPTLossRatio,
     LHOPTTrainingLoss,
     LHOPTValidationLoss,
@@ -56,6 +57,7 @@ __all__ = [
     ModelFamilyOneHot.__name__,
     TrainingLoss.__name__,
     ValidationLoss.__name__,
+    BestObservedValidationLoss.__name__,
     LossRatio.__name__,
     TrainingScore.__name__,
     ValidationScore.__name__,
