@@ -110,3 +110,10 @@ class ModuleTypes(EnumWithIndices):
     Attention = "attention"
     Linear = "linear"
     Embedding = "embedding"
+    LSTM = "lstm"
+
+
+class ToyTaskName(EnumWithIndices):
+
+    XOR = "XOR"
+    NQM = "NQM"
